package ru.dbotthepony.mc.otm.data.world

import com.mojang.serialization.Codec
import com.mojang.serialization.MapCodec
import com.mojang.serialization.codecs.RecordCodecBuilder
import net.minecraft.util.RandomSource
import net.neoforged.bus.api.IEventBus
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.core.ResourceLocation
import ru.dbotthepony.mc.otm.data.codec.minRange
import ru.dbotthepony.mc.otm.registry.MDeferredRegister
import ru.dbotthepony.mc.otm.registry.RegistryDelegate

interface BooleanProvider {
	interface Type<T : BooleanProvider> {
		val codec: MapCodec<T>
	}

	fun instance(): Instance
	val type: Type<*>

	fun interface Instance {
		fun sample(random: RandomSource): Boolean
	}

	class Uniform(val chance: Int) : BooleanProvider, Instance {
		override fun instance(): Instance {
			return this
		}

		override val type: Type<*>
			get() = Companion

		override fun sample(random: RandomSource): Boolean {
			return random.nextInt(chance) == 0
		}

		companion object : Type<Uniform> {
			override val codec: MapCodec<Uniform> = RecordCodecBuilder.mapCodec {
				it.group(Codec.INT.minRange(1).fieldOf("once_every").forGetter(Uniform::chance)).apply(it, ::Uniform)
			}
		}
	}

	companion object {
		private val registryHolder = RegistryDelegate<Type<*>>("decimal_provider_type") {
			defaultKey(ResourceLocation(OverdriveThatMatters.MOD_ID, "zero"))
		}

		val registry by registryHolder
		val registryKey get() = registryHolder.key

		private val registrar = MDeferredRegister(registryKey)

		init {
			registrar.register("uniform") { Uniform.Companion }
		}

		val CODEC: Codec<BooleanProvider> by lazy {
			registry.byNameCodec().dispatch({ it.type }, { it.codec })
		}

		fun register(bus: IEventBus) {
			bus.addListener(registryHolder::build)
			registrar.register(bus)
		}
	}
}
